#
# Proving inequalities
# License: GNU General Public License 3
# Uray M. János
# 2019-2020
#

import time;

load("cad.sage");
load("recur.sage");
load("stopwatch.sage");

def try_to_prove (input, n=None, n0=None, rec=None, cad_alg="builtin", cad_proj=None):
	t = time.time();
	watch = Stopwatch();
	if rec is None:
		if n is None or n0 is None:
			raise ValueError("try_to_prove(): either rec or both n and n0 must be given");
		rec = Recursifier(n, min=n0);
	else:
		n = rec.n_var();
		if n0 is None:
			n0 = rec.min();
	print "Statement:", input;
	watch.start("recursify");
	prop = rec.recursify(logic(input));
	formula = rec.recursify(n >= n0);
	watch.stop();
	print "Substitution:", prop;
	for v in rec.variables():
		print "  ", v, ":", rec.meaning(v);
	print "CAD:", cad_alg;
	ret = None;
	for i in range(4):
		print "\nStep #{}".format(i+1);
		r = 1;
		for k in range(n0, n0 + r):
			watch.start("recursify");
			init = rec.eval(prop, k);
			watch.stop();
			print "in (n={}):".format(k), init;
			init = cylindrical_algebraic_decomposition(init, algorithm=cad_alg, proj=cad_proj, output="poly", watch=watch);
			print "out:", init;
			if init != True:
				ret = False;
				break;
		if ret == False:
			break;
		watch.start("recursify");
		for _ in range(r):
			formula &= prop;
			prop = rec.shift(prop);
		in_formula = logic_implies(formula, prop);
		watch.stop();
		print "in:", in_formula;
		prop = cylindrical_algebraic_decomposition(in_formula, algorithm=cad_alg, output="poly", watch=watch);
		print "out:", prop;
		if prop == True:
			ret = True;
			break;
	print;
	if ret is True:
		print "The statement is proved: TRUE";
	elif ret is False:
		print "The statement is disproved: FALSE";
	else:
		print "Unable to decide";
	t = time.time() - t;
	print t, "s";
	watch.print_all();


var("n,k,x")

# Uncomment exactly one of the examples below.

#try_to_prove(logic_implies(x >= -2, 1 + n*x <= (x + 1)^n), n, 0);

try_to_prove(factorial(n) * sum((-1)^k / factorial(k), k, 0, n) > 0, n, 2);

## 2.4 s
#expr = 1/16^n * (factorial(2*n) / factorial(n)^2)^2;
#try_to_prove(logic_and(1/(4*n) < expr, expr < 1/(3*n + 1)), n, 2);

## 5.3 s
#expr = (1 + n*x^(n+1)) / ((1 + n)*x^n);
#try_to_prove(logic_implies(
	#logic_and(x > 0, x <= 1),
	#logic_and(1 <= expr, expr <= 1 + n/2*(1-x)^2 / x^n)
#), n, 0);

## 47 s
#rec = Recursifier(n);
#function("fibpol", nargs=2);
#rec.add_recurrence([
	#fibpol(n+1, x) == x*fibpol(n, x) + fibpol(n-1, x),
	#fibpol(0, x) == 0,
	#fibpol(1, x) == 1
#]);
#try_to_prove(fibpol(n, x)^2 <= (x^2 + 1)^2 * (x^2 + 2)^(n-3), n, 3, rec=rec, cad_proj="mccallum");

#rec = Recursifier(n);
#function("r", nargs=1);
#rec.add_recurrence([r(n+1) == 1 + n/r(n), r(1) == 1]);
#try_to_prove(logic_and(
	#sqrt(n - 3/4) <= r(n) - 1/2,
	#r(n) - 1/2 <= sqrt(n + 1/4)
#), n, 1, rec=rec);

## 1.66 s
#rec = Recursifier(n, 1);
#function("rooty", nargs=1);
#rec.add_recurrence([rooty(1) == 1, rooty(n+1) == sqrt(n+1 + rooty(n))]);
#try_to_prove(rooty(n) <= sqrt(n) + 1, rec=rec);

#rec = Recursifier(n, 0);
#function("logist", nargs=1);
#var("r");
#rec.add_recurrence(logist(n+1) == r*logist(n)*(1 - logist(n)), min=0);
#try_to_prove(logic_implies(
	#logic_and(r >= 0, r <= 4, logist(0) >= 0, logist(0) <= 1),
	#logic_and(logist(n) >= 0, logist(n) <= 1)
#), rec=rec);

## 2.4 s
#rec = Recursifier(n);
#function("leg", nargs=2);
#rec.add_recurrence([leg(n+2, x) == ((2*n + 3)*x*leg(n+1, x) - (n + 1)*leg(n, x))/(n+2), leg(0, x) == 1, leg(1, x) == x]);
#try_to_prove(logic_implies(
	#logic_and(x >= -1, x <= 1),
	#leg(n+1, x)^2 >= leg(n, x)*leg(n+2, x)
#), rec=rec, cad_proj="mccallum");

## NOTE projection is the slowest
## Cameron T. Cooper, Dr. Fidele F. Ngwane and Dr. Wei-Kai Lai: An improved Fibonacci inequality
#rec = Recursifier(n, 1);
#function("fib", nargs=1);
#rec.add_recurrence([fib(n+1) == fib(n) + fib(n-1), fib(0) == 0, fib(1) == 1]);
#try_to_prove(logic_implies(logic_and(fib(n) > 0, fib(n+1) > 0), 2*(fib(n)^4 + fib(n+1)^4 + fib(n+2)^4)*(1/fib(n)^2 + 1/fib(n+1)^2 + 1/fib(n+2)^2)^2 > 100), rec=rec);

## NOTE builtin too slow
#rec = Recursifier(n, 1);
#function("fib", nargs=1);
#rec.add_recurrence([fib(n+1) == fib(n) + fib(n-1), fib(0) == 0, fib(1) == 1]);
#try_to_prove(sum(sqrt(fib(k)^4 - fib(k)^2 + 1) + (fib(k)^2 - 1)/(fib(k)^4 + 1), k, 1, n, hold=True) < fib(n)*fib(n+1), rec=rec, cad_alg="qepcad");

## NOTE builtin too slow, QEPCAD gives error
#try_to_prove(sum(sqrt(k), k, 1, n)^2 <= sum(k^(1/3), k, 1, n)^3, n, 1, cad_alg="qepcad");

## NOTE builtin too slow
#rec = Recursifier(n, 1);
#function("leg", nargs=2);
#rec.add_recurrence([leg(n+2, x) == ((2*n + 3)*x*leg(n+1, x) - (n + 1)*leg(n, x))/(n+2), leg(0, x) == 1, leg(1, x) == x]);
#try_to_prove(logic_implies(
	#logic_and(x >= 0, x < 1),
	#(n-1)*x*leg(n, x)^2 - (2*n*x^2 + x^2 - 1)*leg(n, x)*leg(n+1, x) + (n+1)*x*leg(n+1, x)^2 >= 0
#), rec=rec, cad_alg="qepcad");

## TODO solve problem with algebraic coefficients
#rec = Recursifier(n, 0);
#function("fib", nargs=1);
#rec.add_recurrence([fib(n+1) == fib(n) + fib(n-1), fib(0) == 0, fib(1) == 1]);
#expr = AA((sqrt(5) + 1)/2)^n / AA(sqrt(5));
#try_to_prove(logic_and(fib(n) > expr - 1/2, fib(n) < expr + 1/2), rec=rec);

## NOTE QEPCAD: lots of erros like:
##    WARNING! Projection factor e^2 i^2 + d i^2 + f^2 + 2 c e f + c^2 e^2
##    is everywhere zero in the cylinder over the cell (4,1,4,1,2,2) of
##    postive dimension.  The McCallum projection may not be valid.
## NOTE QEPCAD: finally:
##    Failure occurred in:    GCSI (final check)
##    Reason for the failure: Too few cells reclaimed.
## NOTE builtin: too slow
#rec = Recursifier(n);
#function("x", nargs=1);
#function("y", nargs=1);
#rec.add_symbolic_function(x);
#rec.add_symbolic_function(y);
#try_to_prove(logic_implies(
	#logic_and(sum(x(k)^2, k, 1, n) > 0, sum(y(k)^2, k, 1, n) > 0),
	#sum(x(k)*y(k), k, 1, n)^2 <= sum(x(k)^2, k, 1, n) * sum(y(k)^2, k, 1, n)
#), rec=rec);

