#
# rootof()-expression
# License: GNU General Public License 3
# Uray M. János
# 2019-2020
#

#
# Define the symbolic expression rootof(p, x, i), where:
#    p: a (possibly multivariate) polynomial,
#    x: a variable name in p,
#    i: the root index (between 0 <= i < n).
# The expression rootof(p, x, i) means the 'i'th root of p in the variable x.
# This is a function of the other variables in p. This may be simplified for
# simple cases, e.g. for univariate, linear or quadratic polynomials.
#
def eval_rootof (self, p, x, i):
	if not x.is_symbol():
		raise ValueError("rootof(): " + str(x) + " is not a symbolic variable");
	p = SR(p);
	if not p.is_polynomial(x):
		raise ValueError("rootof(): " + str(p) + " is not a polynomial in " + str(x));
	p = p.polynomial(ring=SR[x]);
	x = p.parent().gen();
	if p.is_constant():
		if p.is_zero():
			return SR(0);
		else:
			raise ValueError("rootof(): " + str(p) + " has no roots");
	n = p.degree();
	if i < 0 or i >= n:
		raise ValueError("rootof(): root index " + str(i) + " is invalid");
	if n == 1:
		b,a = p.list();
		return -b / a;
	if n == 2:
		c,b,a = p.list();
		asign = None;
		if a.is_positive():
			asign = +1;
		elif a.is_negative():
			asign = -1;
		if asign != None:
			d = b^2 - 4*a*c;
			if d.is_negative() or (d.is_zero() and i == 1):
				raise ValueError("rootof(): root index " + str(i) + " is invalid");
			if isinstance(d, Expression):
				d = d.expand();
			if i == 0:
				return (-b - sqrt(d))/(2*a);
			else:
				return (-b + sqrt(d))/(2*a);
	if all(c.is_constant() for c in p):
		roots = p.roots(AA, multiplicities=False);
		if i >= len(roots):
			raise ValueError("rootof(): root index " + str(i) + " is invalid");
		roots.sort();
		root = roots[i];
		if root in QQ:
			return QQ(root);
		else:
			return root;
	pass;

function("rootof", nargs=3, eval_func=eval_rootof);

