#
# Multi-label stopwatch
# License: GNU General Public License 3
# Uray M. János
# 2019-2020
#

import time;

class Stopwatch:
	"""
	A class to measure and store time with different labels.
	"""
	
	def __init__ (self):
		self._times = dict();
		self._names = [];
		self._current = None;
		self._start = None;
	
	def start (self, name):
		"""
		Start measuring the time under the label 'name'.
		If currently other timing is active, it is stopped first, i.e. the class
		cannot measure for multiple labels at once.
		If the same label was already used, the time is added to the previous time.
		"""
		if self._current:
			self.stop();
		self._current = name;
		self._start = time.time();
	
	def stop (self):
		"""
		Stop measuring the time.
		If no measuring is active, nothing happens.
		"""
		if self._current:
			dur = time.time() - self._start;
			if self._current in self._times:
				self._times[self._current] += dur;
			else:
				self._times[self._current] = dur;
				self._names.append(self._current);
			self._current = None;
			self._start = None;
	
	def get (self, name):
		"""
		Return the time measured so far for the given label.
		It must be called in a stopped state for correct result.
		"""
		return self._times[name];
	
	def print_all (self):
		"""
		Print all times for all labels.
		"""
		max_len = max(len(name) for name in self._names) if self._names else None;
		for name in self._names:
			print name + ":" + " " * (max_len - len(name)), "{:2.3f} s".format(self._times[name]);

